"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
There is an inconsistency in behavior
of current QuerySet.get_or_create():
It avoids unnecessary DB query for related object given as argument
only if created is True.
I gave a script demonstrating this here:
https://code.djangoproject.com/ticket/34884
The patch of get_or_create() can be replaced by
a more general patch on QuerySet.get().
"""

# pylint: disable-next=import-error
from django.db.models import QuerySet

from .django__query_set import put_filter_arg_in_field_cache

original_get = QuerySet.get


def patched_get_v1(self, *args, **kwargs):
    """
    Patch for QuerySet.get() with foreign key cache.
    """
    result = original_get(self, *args, **kwargs)
    put_filter_arg_in_field_cache(result, kwargs)
    return result


def apply_patched_get_v1():
    """
    Apply the patch for QuerySet.get() with foreign key cache.
    """
    QuerySet.get = patched_get_v1
    return original_get



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
"""

# pylint: disable-next=import-error
from django.core.exceptions import FieldDoesNotExist

# pylint: disable-next=import-error
from django.db.models import ForeignKey


def put_filter_arg_in_field_cache(obj, kwargs):
    """
    Given an object and a dict of field-values/filtering-args
    put in the cache of ForeignKey/OneToOneField the given objects.
    """
    for key, value in kwargs.items():
        try:
            # pylint: disable-next=protected-access
            if isinstance(obj._meta.get_field(key), ForeignKey):
                # isinstance handles OneToOneField also.
                setattr(obj, key, value)
        except FieldDoesNotExist:
            pass



r"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
Choose between flaky tests by default,
and slow tests by default.
The topic is complex,
because you may end up having tests
where most of the time is spent
on the same cache filling over and over.
But the good news is that you can apply or not this patch at will.

/!\/!\/!\ cache.clear() can do a lot of harm
if you have a cache infrastructure shared
between your test environment and your production environment,
or shared between your Django application and another application.
As the Django doc states at
https://docs.djangoproject.com/en/dev/topics/cache/:
> Finally, if you want to delete all the keys in the cache,
> use cache.clear().
> Be careful with this; clear() will remove everything from the cache,
> not just the keys set by your application.
Better solutions exist by searching for cache keys with some prefix,
and then using cache.delete_many().
But there is no uniform way yet, across cache backends,
to get cache keys with some prefix.
See:
https://stackoverflow.com/questions/9048257/\
  get-list-of-cache-keys-in-django
For this solution and Redis,
you will have to choose v2 and modify your caches settings
according to your use case.

/!\ Make sure you call super().tearDown()
in your custom tearDown() methods.
"""

import os
from typing import Dict

# pylint: disable-next=import-error
from django.conf import settings

# pylint: disable-next=import-error
from django.core.cache import cache, caches

# pylint: disable-next=import-error
from django.test import TestCase

from .django__base_cache__make_cache_key import (
    make_parallel_test_cache_key,
    make_test_cache_key,
)

# This function is a noop (pass).
original_tear_down = TestCase.tearDown


# pylint: disable-next=unused-argument
def patched_tear_down_v1(self, *args, **kwargs):
    """
    Not recommended,
    it will wipe out everything in the current cache.
    """
    cache.clear()


def apply_patched_tear_down_v1():
    """
    Apply a not recommended patch,
    it will wipe out everything in the current cache.
    """
    TestCase.tearDown = patched_tear_down_v1


# pylint: disable-next=unused-argument
def patched_tear_down_v2(self, *args, **kwargs):
    """
    This is the most flexible patch
    where you can define on a per cache basis
    the function you want to apply to clear the cache.
    The code of 4 possible callback functions is given after.
    """
    for cache_alias, cache_settings in settings.CACHES.items():
        test_clear_cache_callback = cache_settings.get(
            "TEST_CLEAR_CACHE_CALLBACK",
        )
        if test_clear_cache_callback is not None:
            test_clear_cache_callback(
                cache_alias,
                cache_settings,
                caches[cache_alias],
            )


def apply_patched_tear_down_v2():
    """
    Apply the recommended and flexible patch
    for cleaning cache in tests.
    """
    TestCase.tearDown = patched_tear_down_v2


def test_clear_cache_callback_v1(
    # pylint: disable-next=unused-argument
    cache_alias: str,
    # pylint: disable-next=unused-argument
    cache_settings: Dict,
    current_cache,
):
    """Probably not the best option"""
    current_cache.clear()


def test_clear_cache_callback_v2(
    # pylint: disable-next=unused-argument
    cache_alias: str,
    cache_settings: Dict,
    current_cache,
):
    """
    A nice possibility with Redis.
    See https://stackoverflow.com/a/55633767/5796086
    CC BY SA 4.0
    and see make_test_cache_key(), make_parallel_test_cache_key()
    in this package.
    And see v5...
    in case .keys() and .delete_many() apply KEY_FUNCTION
    and you want to cover all versions
    (the third parameter of KEY_FUNCTION)...
    """
    key_function = cache_settings.get("KEY_FUNCTION")
    key_prefix = cache_settings.get("KEY_PREFIX", "")
    if key_function is make_test_cache_key:
        query_key_prefix = f"test_cache:{key_prefix}:*"
    elif key_function is make_parallel_test_cache_key:
        query_key_prefix = f"test_cache{os.getpid()}:{key_prefix}:*"
    else:
        return
    cache_keys = current_cache.keys(query_key_prefix)
    if len(cache_keys) > 0:
        current_cache.delete_many(cache_keys)


def test_clear_cache_callback_v3(
    # pylint: disable-next=unused-argument
    cache_alias: str,
    cache_settings: Dict,
    current_cache,
):
    """
    Another possibility with other versions of Redis.
    See https://stackoverflow.com/a/71698532/5796086
    CC BY SA 4.0
    and see make_test_cache_key(), make_parallel_test_cache_key()
    in this package.
    """
    key_function = cache_settings.get("KEY_FUNCTION")
    key_prefix = cache_settings.get("KEY_PREFIX", "")
    if key_function is make_test_cache_key:
        query_key_prefix = f"test_cache:{key_prefix}:*"
    elif key_function is make_parallel_test_cache_key:
        query_key_prefix = f"test_cache{os.getpid()}:{key_prefix}:*"
    else:
        return
    cache_keys = current_cache.get_client(1).keys(query_key_prefix)
    if len(cache_keys) > 0:
        current_cache.delete_many(cache_keys)


def test_clear_cache_callback_v4(
    # pylint: disable-next=unused-argument
    cache_alias: str,
    cache_settings: Dict,
    current_cache,
):
    """
    Still another possibility with other versions of Redis.
    See https://stackoverflow.com/a/71698532/5796086
    CC BY SA 4.0
    and see make_test_cache_key(), make_parallel_test_cache_key()
    in this package.
    """
    key_function = cache_settings.get("KEY_FUNCTION")
    key_prefix = cache_settings.get("KEY_PREFIX", "")
    if key_function is make_test_cache_key:
        query_key_prefix = f"test_cache:{key_prefix}:*"
    elif key_function is make_parallel_test_cache_key:
        query_key_prefix = f"test_cache{os.getpid()}:{key_prefix}:*"
    else:
        return
    cache_keys = current_cache.get_client().keys(query_key_prefix)
    if len(cache_keys) > 0:
        current_cache.delete_many(cache_keys)


def test_clear_cache_callback_v5(
    # pylint: disable-next=unused-argument
    cache_alias: str,
    cache_settings: Dict,
    current_cache,
):
    """
    A nice possibility with Redis.
    """
    key_function = cache_settings.get("KEY_FUNCTION")
    key_prefix = cache_settings.get("KEY_PREFIX", "")
    if key_function is make_test_cache_key:
        query_key_prefix = f"test_cache:{key_prefix}:*"
    elif key_function is make_parallel_test_cache_key:
        query_key_prefix = f"test_cache{os.getpid()}:{key_prefix}:*"
    else:
        return
    cache_keys = current_cache.client.get_client(write=False).keys(
        query_key_prefix
    )
    if len(cache_keys) > 0:
        current_cache.client.get_client(write=True).delete(
            *cache_keys
        )



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
The goal of dev_side_code files is to keep code
that should not be used
but was used to create the best code possible in some patches.
Below, is the code for micro-optimisations of
retrieve_forward_cache_callback_for_foreign_key.
"""

# pylint: disable=too-many-lines

import functools
import random
import timeit

# pylint: disable-next=invalid-name
instance_with_fk = None
# pylint: disable-next=invalid-name
instance_without_fk = None
e = ()  # empty_tuple


# l = lambda
# c = callback
# cc = create_callback
# Dict of callbacks to test
callbacks = {
    "l_field_list_list": lambda x: (
        [x.field_name] if x.field_name else []
    ),
    "l_field_id_list_list": lambda x: (
        [x.field_name] if x.field_name_id else []
    ),
    "l_field_not_none_list_list": lambda x: (
        [x.field_name] if x.field_name is not None else []
    ),
    "l_field_id_not_none_list_list": lambda x: (
        [x.field_name] if x.field_name_id is not None else []
    ),
    "l_field_none_list_list": lambda x: (
        [] if x.field_name is None else [x.field_name]
    ),
    "l_field_id_none_list_list": lambda x: (
        [] if x.field_name_id is None else [x.field_name]
    ),
    "l_field_tuple_tuple": lambda x: (
        (x.field_name,) if x.field_name else ()
    ),
    "l_field_id_tuple_tuple": lambda x: (
        (x.field_name,) if x.field_name_id else ()
    ),
    "l_field_not_none_tuple_tuple": lambda x: (
        (x.field_name,) if x.field_name is not None else ()
    ),
    "l_field_id_not_none_tuple_tuple": lambda x: (
        (x.field_name,) if x.field_name_id is not None else ()
    ),
    "l_field_none_tuple_tuple": lambda x: (
        () if x.field_name is None else (x.field_name,)
    ),
    "l_field_id_none_tuple_tuple": lambda x: (
        () if x.field_name_id is None else (x.field_name,)
    ),
    "l_field_list_tuple": lambda x: (
        [x.field_name] if x.field_name else ()
    ),
    "l_field_id_list_tuple": lambda x: (
        [x.field_name] if x.field_name_id else ()
    ),
    "l_field_not_none_list_tuple": lambda x: (
        [x.field_name] if x.field_name is not None else ()
    ),
    "l_field_id_not_none_list_tuple": lambda x: (
        [x.field_name] if x.field_name_id is not None else ()
    ),
    "l_field_none_list_tuple": lambda x: (
        () if x.field_name is None else [x.field_name]
    ),
    "l_field_id_none_list_tuple": lambda x: (
        () if x.field_name_id is None else [x.field_name]
    ),
    "l_field_tuple_list": lambda x: (
        (x.field_name,) if x.field_name else []
    ),
    "l_field_id_tuple_list": lambda x: (
        (x.field_name,) if x.field_name_id else []
    ),
    "l_field_not_none_tuple_list": lambda x: (
        (x.field_name,) if x.field_name is not None else []
    ),
    "l_field_id_not_none_tuple_list": lambda x: (
        (x.field_name,) if x.field_name_id is not None else []
    ),
    "l_field_none_tuple_list": lambda x: (
        [] if x.field_name is None else (x.field_name,)
    ),
    "l_field_id_none_tuple_list": lambda x: (
        [] if x.field_name_id is None else (x.field_name,)
    ),
    "l_field_tuple_empty": lambda x: (
        (x.field_name,) if x.field_name else e
    ),
    "l_field_id_tuple_empty": lambda x: (
        (x.field_name,) if x.field_name_id else e
    ),
    "l_field_not_none_tuple_empty": lambda x: (
        (x.field_name,) if x.field_name is not None else e
    ),
    "l_field_id_not_none_tuple_empty": lambda x: (
        (x.field_name,) if x.field_name_id is not None else e
    ),
    "l_field_none_tuple_empty": lambda x: (
        e if x.field_name is None else (x.field_name,)
    ),
    "l_field_id_none_tuple_empty": lambda x: (
        e if x.field_name_id is None else (x.field_name,)
    ),
    "l_field_list_empty": lambda x: (
        [x.field_name] if x.field_name else e
    ),
    "l_field_id_list_empty": lambda x: (
        [x.field_name] if x.field_name_id else e
    ),
    "l_field_not_none_list_empty": lambda x: (
        [x.field_name] if x.field_name is not None else e
    ),
    "l_field_id_not_none_list_empty": lambda x: (
        [x.field_name] if x.field_name_id is not None else e
    ),
    "l_field_none_list_empty": lambda x: (
        e if x.field_name is None else [x.field_name]
    ),
    "l_field_id_none_list_empty": lambda x: (
        e if x.field_name_id is None else [x.field_name]
    ),
}


def cc_field_list_list(f):
    """
    A function to obtain a c_field_list_list.
    """

    def c_field_list_list(x):
        """
        The customized c_field_list_list.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) else []

    return c_field_list_list


def cc_field_id_list_list(f):
    """
    A function to obtain a c_field_id_list_list.
    """
    g = f + "_id"

    def c_field_id_list_list(x):
        """
        The customized c_field_id_list_list.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) else []

    return c_field_id_list_list


def cc_field_not_none_list_list(f):
    """
    A function to obtain a c_field_not_none_list_list.
    """

    def c_field_not_none_list_list(x):
        """
        The customized c_field_not_none_list_list.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) is not None else []

    return c_field_not_none_list_list


def cc_field_id_not_none_list_list(f):
    """
    A function to obtain a c_field_id_not_none_list_list.
    """
    g = f + "_id"

    def c_field_id_not_none_list_list(x):
        """
        The customized c_field_id_not_none_list_list.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) is not None else []

    return c_field_id_not_none_list_list


def cc_field_none_list_list(f):
    """
    A function to obtain a c_field_none_list_list.
    """

    def c_field_none_list_list(x):
        """
        The customized c_field_none_list_list.
        """
        nonlocal f

        return [] if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_list


def cc_field_id_none_list_list(f):
    """
    A function to obtain a c_field_id_none_list_list.
    """
    g = f + "_id"

    def c_field_id_none_list_list(x):
        """
        The customized c_field_id_none_list_list.
        """
        nonlocal f
        nonlocal g

        return [] if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_list


def cc_field_tuple_tuple(f):
    """
    A function to obtain a c_field_tuple_tuple.
    """

    def c_field_tuple_tuple(x):
        """
        The customized c_field_tuple_tuple.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) else ()

    return c_field_tuple_tuple


def cc_field_id_tuple_tuple(f):
    """
    A function to obtain a c_field_id_tuple_tuple.
    """
    g = f + "_id"

    def c_field_id_tuple_tuple(x):
        """
        The customized c_field_id_tuple_tuple.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) else ()

    return c_field_id_tuple_tuple


def cc_field_not_none_tuple_tuple(f):
    """
    A function to obtain a c_field_not_none_tuple_tuple.
    """

    def c_field_not_none_tuple_tuple(x):
        """
        The customized c_field_not_none_tuple_tuple.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) is not None else ()

    return c_field_not_none_tuple_tuple


def cc_field_id_not_none_tuple_tuple(f):
    """
    A function to obtain a c_field_id_not_none_tuple_tuple.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_tuple(x):
        """
        The customized c_field_id_not_none_tuple_tuple.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) is not None else ()

    return c_field_id_not_none_tuple_tuple


def cc_field_none_tuple_tuple(f):
    """
    A function to obtain a c_field_none_tuple_tuple.
    """

    def c_field_none_tuple_tuple(x):
        """
        The customized c_field_none_tuple_tuple.
        """
        nonlocal f

        return () if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_tuple


def cc_field_id_none_tuple_tuple(f):
    """
    A function to obtain a c_field_id_none_tuple_tuple.
    """
    g = f + "_id"

    def c_field_id_none_tuple_tuple(x):
        """
        The customized c_field_id_none_tuple_tuple.
        """
        nonlocal f
        nonlocal g

        return () if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_tuple


def cc_field_list_tuple(f):
    """
    A function to obtain a c_field_list_tuple.
    """

    def c_field_list_tuple(x):
        """
        The customized c_field_list_tuple.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) else ()

    return c_field_list_tuple


def cc_field_id_list_tuple(f):
    """
    A function to obtain a c_field_id_list_tuple.
    """
    g = f + "_id"

    def c_field_id_list_tuple(x):
        """
        The customized c_field_id_list_tuple.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) else ()

    return c_field_id_list_tuple


def cc_field_not_none_list_tuple(f):
    """
    A function to obtain a c_field_not_none_list_tuple.
    """

    def c_field_not_none_list_tuple(x):
        """
        The customized c_field_not_none_list_tuple.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) is not None else ()

    return c_field_not_none_list_tuple


def cc_field_id_not_none_list_tuple(f):
    """
    A function to obtain a c_field_id_not_none_list_tuple.
    """
    g = f + "_id"

    def c_field_id_not_none_list_tuple(x):
        """
        The customized c_field_id_not_none_list_tuple.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) is not None else ()

    return c_field_id_not_none_list_tuple


def cc_field_none_list_tuple(f):
    """
    A function to obtain a c_field_none_list_tuple.
    """

    def c_field_none_list_tuple(x):
        """
        The customized c_field_none_list_tuple.
        """
        nonlocal f

        return () if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_tuple


def cc_field_id_none_list_tuple(f):
    """
    A function to obtain a c_field_id_none_list_tuple.
    """
    g = f + "_id"

    def c_field_id_none_list_tuple(x):
        """
        The customized c_field_id_none_list_tuple.
        """
        nonlocal f
        nonlocal g

        return () if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_tuple


def cc_field_tuple_list(f):
    """
    A function to obtain a c_field_tuple_list.
    """

    def c_field_tuple_list(x):
        """
        The customized c_field_tuple_list.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) else []

    return c_field_tuple_list


def cc_field_id_tuple_list(f):
    """
    A function to obtain a c_field_id_tuple_list.
    """
    g = f + "_id"

    def c_field_id_tuple_list(x):
        """
        The customized c_field_id_tuple_list.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) else []

    return c_field_id_tuple_list


def cc_field_not_none_tuple_list(f):
    """
    A function to obtain a c_field_not_none_tuple_list.
    """

    def c_field_not_none_tuple_list(x):
        """
        The customized c_field_not_none_tuple_list.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) is not None else []

    return c_field_not_none_tuple_list


def cc_field_id_not_none_tuple_list(f):
    """
    A function to obtain a c_field_id_not_none_tuple_list.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_list(x):
        """
        The customized c_field_id_not_none_tuple_list.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) is not None else []

    return c_field_id_not_none_tuple_list


def cc_field_none_tuple_list(f):
    """
    A function to obtain a c_field_none_tuple_list.
    """

    def c_field_none_tuple_list(x):
        """
        The customized c_field_none_tuple_list.
        """
        nonlocal f

        return [] if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_list


def cc_field_id_none_tuple_list(f):
    """
    A function to obtain a c_field_id_none_tuple_list.
    """
    g = f + "_id"

    def c_field_id_none_tuple_list(x):
        """
        The customized c_field_id_none_tuple_list.
        """
        nonlocal f
        nonlocal g

        return [] if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_list


def cc_field_tuple_empty(f):
    """
    A function to obtain a c_field_tuple_empty.
    """

    def c_field_tuple_empty(x):
        """
        The customized c_field_tuple_empty.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) else e

    return c_field_tuple_empty


def cc_field_id_tuple_empty(f):
    """
    A function to obtain a c_field_id_tuple_empty.
    """
    g = f + "_id"

    def c_field_id_tuple_empty(x):
        """
        The customized c_field_id_tuple_empty.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) else e

    return c_field_id_tuple_empty


def cc_field_not_none_tuple_empty(f):
    """
    A function to obtain a c_field_not_none_tuple_empty.
    """

    def c_field_not_none_tuple_empty(x):
        """
        The customized c_field_not_none_tuple_empty.
        """
        nonlocal f

        return (getattr(x, f),) if getattr(x, f) is not None else e

    return c_field_not_none_tuple_empty


def cc_field_id_not_none_tuple_empty(f):
    """
    A function to obtain a c_field_id_not_none_tuple_empty.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_empty(x):
        """
        The customized c_field_id_not_none_tuple_empty.
        """
        nonlocal f
        nonlocal g

        return (getattr(x, f),) if getattr(x, g) is not None else e

    return c_field_id_not_none_tuple_empty


def cc_field_none_tuple_empty(f):
    """
    A function to obtain a c_field_none_tuple_empty.
    """

    def c_field_none_tuple_empty(x):
        """
        The customized c_field_none_tuple_empty.
        """
        nonlocal f

        return e if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_empty


def cc_field_id_none_tuple_empty(f):
    """
    A function to obtain a c_field_id_none_tuple_empty.
    """
    g = f + "_id"

    def c_field_id_none_tuple_empty(x):
        """
        The customized c_field_id_none_tuple_empty.
        """
        nonlocal f
        nonlocal g

        return e if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_empty


def cc_field_list_empty(f):
    """
    A function to obtain a c_field_list_empty.
    """

    def c_field_list_empty(x):
        """
        The customized c_field_list_empty.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) else e

    return c_field_list_empty


def cc_field_id_list_empty(f):
    """
    A function to obtain a c_field_id_list_empty.
    """
    g = f + "_id"

    def c_field_id_list_empty(x):
        """
        The customized c_field_id_list_empty.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) else e

    return c_field_id_list_empty


def cc_field_not_none_list_empty(f):
    """
    A function to obtain a c_field_not_none_list_empty.
    """

    def c_field_not_none_list_empty(x):
        """
        The customized c_field_not_none_list_empty.
        """
        nonlocal f

        return [getattr(x, f)] if getattr(x, f) is not None else e

    return c_field_not_none_list_empty


def cc_field_id_not_none_list_empty(f):
    """
    A function to obtain a c_field_id_not_none_list_empty.
    """
    g = f + "_id"

    def c_field_id_not_none_list_empty(x):
        """
        The customized c_field_id_not_none_list_empty.
        """
        nonlocal f
        nonlocal g

        return [getattr(x, f)] if getattr(x, g) is not None else e

    return c_field_id_not_none_list_empty


def cc_field_none_list_empty(f):
    """
    A function to obtain a c_field_none_list_empty.
    """

    def c_field_none_list_empty(x):
        """
        The customized c_field_none_list_empty.
        """
        nonlocal f

        return e if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_empty


def cc_field_id_none_list_empty(f):
    """
    A function to obtain a c_field_id_none_list_empty.
    """
    g = f + "_id"

    def c_field_id_none_list_empty(x):
        """
        The customized c_field_id_none_list_empty.
        """
        nonlocal f
        nonlocal g

        return e if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_empty


def cc_field_list_list_nnl(f):
    """
    A function to obtain a c_field_list_list_nnl.
    """

    def c_field_list_list_nnl(x):
        """
        The customized c_field_list_list_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) else []

    return c_field_list_list_nnl


def cc_field_id_list_list_nnl(f):
    """
    A function to obtain a c_field_id_list_list_nnl.
    """
    g = f + "_id"

    def c_field_id_list_list_nnl(x):
        """
        The customized c_field_id_list_list_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) else []

    return c_field_id_list_list_nnl


def cc_field_not_none_list_list_nnl(f):
    """
    A function to obtain a c_field_not_none_list_list_nnl.
    """

    def c_field_not_none_list_list_nnl(x):
        """
        The customized c_field_not_none_list_list_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) is not None else []

    return c_field_not_none_list_list_nnl


def cc_field_id_not_none_list_list_nnl(f):
    """
    A function to obtain a c_field_id_not_none_list_list_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_list_list_nnl(x):
        """
        The customized c_field_id_not_none_list_list_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) is not None else []

    return c_field_id_not_none_list_list_nnl


def cc_field_none_list_list_nnl(f):
    """
    A function to obtain a c_field_none_list_list_nnl.
    """

    def c_field_none_list_list_nnl(x):
        """
        The customized c_field_none_list_list_nnl.
        """

        return [] if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_list_nnl


def cc_field_id_none_list_list_nnl(f):
    """
    A function to obtain a c_field_id_none_list_list_nnl.
    """
    g = f + "_id"

    def c_field_id_none_list_list_nnl(x):
        """
        The customized c_field_id_none_list_list_nnl.
        """

        return [] if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_list_nnl


def cc_field_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_tuple_tuple_nnl.
    """

    def c_field_tuple_tuple_nnl(x):
        """
        The customized c_field_tuple_tuple_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) else ()

    return c_field_tuple_tuple_nnl


def cc_field_id_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_id_tuple_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_tuple_tuple_nnl(x):
        """
        The customized c_field_id_tuple_tuple_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) else ()

    return c_field_id_tuple_tuple_nnl


def cc_field_not_none_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_not_none_tuple_tuple_nnl.
    """

    def c_field_not_none_tuple_tuple_nnl(x):
        """
        The customized c_field_not_none_tuple_tuple_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) is not None else ()

    return c_field_not_none_tuple_tuple_nnl


def cc_field_id_not_none_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_id_not_none_tuple_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_tuple_nnl(x):
        """
        The customized c_field_id_not_none_tuple_tuple_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) is not None else ()

    return c_field_id_not_none_tuple_tuple_nnl


def cc_field_none_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_none_tuple_tuple_nnl.
    """

    def c_field_none_tuple_tuple_nnl(x):
        """
        The customized c_field_none_tuple_tuple_nnl.
        """

        return () if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_tuple_nnl


def cc_field_id_none_tuple_tuple_nnl(f):
    """
    A function to obtain a c_field_id_none_tuple_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_none_tuple_tuple_nnl(x):
        """
        The customized c_field_id_none_tuple_tuple_nnl.
        """

        return () if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_tuple_nnl


def cc_field_list_tuple_nnl(f):
    """
    A function to obtain a c_field_list_tuple_nnl.
    """

    def c_field_list_tuple_nnl(x):
        """
        The customized c_field_list_tuple_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) else ()

    return c_field_list_tuple_nnl


def cc_field_id_list_tuple_nnl(f):
    """
    A function to obtain a c_field_id_list_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_list_tuple_nnl(x):
        """
        The customized c_field_id_list_tuple_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) else ()

    return c_field_id_list_tuple_nnl


def cc_field_not_none_list_tuple_nnl(f):
    """
    A function to obtain a c_field_not_none_list_tuple_nnl.
    """

    def c_field_not_none_list_tuple_nnl(x):
        """
        The customized c_field_not_none_list_tuple_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) is not None else ()

    return c_field_not_none_list_tuple_nnl


def cc_field_id_not_none_list_tuple_nnl(f):
    """
    A function to obtain a c_field_id_not_none_list_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_list_tuple_nnl(x):
        """
        The customized c_field_id_not_none_list_tuple_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) is not None else ()

    return c_field_id_not_none_list_tuple_nnl


def cc_field_none_list_tuple_nnl(f):
    """
    A function to obtain a c_field_none_list_tuple_nnl.
    """

    def c_field_none_list_tuple_nnl(x):
        """
        The customized c_field_none_list_tuple_nnl.
        """

        return () if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_tuple_nnl


def cc_field_id_none_list_tuple_nnl(f):
    """
    A function to obtain a c_field_id_none_list_tuple_nnl.
    """
    g = f + "_id"

    def c_field_id_none_list_tuple_nnl(x):
        """
        The customized c_field_id_none_list_tuple_nnl.
        """

        return () if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_tuple_nnl


def cc_field_tuple_list_nnl(f):
    """
    A function to obtain a c_field_tuple_list_nnl.
    """

    def c_field_tuple_list_nnl(x):
        """
        The customized c_field_tuple_list_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) else []

    return c_field_tuple_list_nnl


def cc_field_id_tuple_list_nnl(f):
    """
    A function to obtain a c_field_id_tuple_list_nnl.
    """
    g = f + "_id"

    def c_field_id_tuple_list_nnl(x):
        """
        The customized c_field_id_tuple_list_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) else []

    return c_field_id_tuple_list_nnl


def cc_field_not_none_tuple_list_nnl(f):
    """
    A function to obtain a c_field_not_none_tuple_list_nnl.
    """

    def c_field_not_none_tuple_list_nnl(x):
        """
        The customized c_field_not_none_tuple_list_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) is not None else []

    return c_field_not_none_tuple_list_nnl


def cc_field_id_not_none_tuple_list_nnl(f):
    """
    A function to obtain a c_field_id_not_none_tuple_list_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_list_nnl(x):
        """
        The customized c_field_id_not_none_tuple_list_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) is not None else []

    return c_field_id_not_none_tuple_list_nnl


def cc_field_none_tuple_list_nnl(f):
    """
    A function to obtain a c_field_none_tuple_list_nnl.
    """

    def c_field_none_tuple_list_nnl(x):
        """
        The customized c_field_none_tuple_list_nnl.
        """

        return [] if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_list_nnl


def cc_field_id_none_tuple_list_nnl(f):
    """
    A function to obtain a c_field_id_none_tuple_list_nnl.
    """
    g = f + "_id"

    def c_field_id_none_tuple_list_nnl(x):
        """
        The customized c_field_id_none_tuple_list_nnl.
        """

        return [] if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_list_nnl


def cc_field_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_tuple_empty_nnl.
    """

    def c_field_tuple_empty_nnl(x):
        """
        The customized c_field_tuple_empty_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) else e

    return c_field_tuple_empty_nnl


def cc_field_id_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_id_tuple_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_tuple_empty_nnl(x):
        """
        The customized c_field_id_tuple_empty_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) else e

    return c_field_id_tuple_empty_nnl


def cc_field_not_none_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_not_none_tuple_empty_nnl.
    """

    def c_field_not_none_tuple_empty_nnl(x):
        """
        The customized c_field_not_none_tuple_empty_nnl.
        """

        return (getattr(x, f),) if getattr(x, f) is not None else e

    return c_field_not_none_tuple_empty_nnl


def cc_field_id_not_none_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_id_not_none_tuple_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_tuple_empty_nnl(x):
        """
        The customized c_field_id_not_none_tuple_empty_nnl.
        """

        return (getattr(x, f),) if getattr(x, g) is not None else e

    return c_field_id_not_none_tuple_empty_nnl


def cc_field_none_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_none_tuple_empty_nnl.
    """

    def c_field_none_tuple_empty_nnl(x):
        """
        The customized c_field_none_tuple_empty_nnl.
        """

        return e if getattr(x, f) is None else (getattr(x, f),)

    return c_field_none_tuple_empty_nnl


def cc_field_id_none_tuple_empty_nnl(f):
    """
    A function to obtain a c_field_id_none_tuple_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_none_tuple_empty_nnl(x):
        """
        The customized c_field_id_none_tuple_empty_nnl.
        """

        return e if getattr(x, g) is None else (getattr(x, f),)

    return c_field_id_none_tuple_empty_nnl


def cc_field_list_empty_nnl(f):
    """
    A function to obtain a c_field_list_empty_nnl.
    """

    def c_field_list_empty_nnl(x):
        """
        The customized c_field_list_empty_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) else e

    return c_field_list_empty_nnl


def cc_field_id_list_empty_nnl(f):
    """
    A function to obtain a c_field_id_list_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_list_empty_nnl(x):
        """
        The customized c_field_id_list_empty_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) else e

    return c_field_id_list_empty_nnl


def cc_field_not_none_list_empty_nnl(f):
    """
    A function to obtain a c_field_not_none_list_empty_nnl.
    """

    def c_field_not_none_list_empty_nnl(x):
        """
        The customized c_field_not_none_list_empty_nnl.
        """

        return [getattr(x, f)] if getattr(x, f) is not None else e

    return c_field_not_none_list_empty_nnl


def cc_field_id_not_none_list_empty_nnl(f):
    """
    A function to obtain a c_field_id_not_none_list_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_not_none_list_empty_nnl(x):
        """
        The customized c_field_id_not_none_list_empty_nnl.
        """

        return [getattr(x, f)] if getattr(x, g) is not None else e

    return c_field_id_not_none_list_empty_nnl


def cc_field_none_list_empty_nnl(f):
    """
    A function to obtain a c_field_none_list_empty_nnl.
    """

    def c_field_none_list_empty_nnl(x):
        """
        The customized c_field_none_list_empty_nnl.
        """

        return e if getattr(x, f) is None else [getattr(x, f)]

    return c_field_none_list_empty_nnl


def cc_field_id_none_list_empty_nnl(f):
    """
    A function to obtain a c_field_id_none_list_empty_nnl.
    """
    g = f + "_id"

    def c_field_id_none_list_empty_nnl(x):
        """
        The customized c_field_id_none_list_empty_nnl.
        """

        return e if getattr(x, g) is None else [getattr(x, f)]

    return c_field_id_none_list_empty_nnl


callbacks.update(
    {
        "c_field_list_list": cc_field_list_list("field_name"),
        "c_field_id_list_list": cc_field_id_list_list("field_name"),
        "c_field_not_none_list_list": cc_field_not_none_list_list(
            "field_name"
        ),
        "c_field_id_not_none_list_list": (
            cc_field_id_not_none_list_list("field_name")
        ),
        "c_field_none_list_list": cc_field_none_list_list(
            "field_name"
        ),
        "c_field_id_none_list_list": cc_field_id_none_list_list(
            "field_name"
        ),
        "c_field_tuple_tuple": cc_field_tuple_tuple("field_name"),
        "c_field_id_tuple_tuple": cc_field_id_tuple_tuple(
            "field_name"
        ),
        "c_field_not_none_tuple_tuple": cc_field_not_none_tuple_tuple(
            "field_name"
        ),
        "c_field_id_not_none_tuple_tuple": (
            cc_field_id_not_none_tuple_tuple("field_name")
        ),
        "c_field_none_tuple_tuple": cc_field_none_tuple_tuple(
            "field_name"
        ),
        "c_field_id_none_tuple_tuple": cc_field_id_none_tuple_tuple(
            "field_name"
        ),
        "c_field_list_tuple": cc_field_list_tuple("field_name"),
        "c_field_id_list_tuple": cc_field_id_list_tuple("field_name"),
        "c_field_not_none_list_tuple": cc_field_not_none_list_tuple(
            "field_name"
        ),
        "c_field_id_not_none_list_tuple": (
            cc_field_id_not_none_list_tuple("field_name")
        ),
        "c_field_none_list_tuple": cc_field_none_list_tuple(
            "field_name"
        ),
        "c_field_id_none_list_tuple": cc_field_id_none_list_tuple(
            "field_name"
        ),
        "c_field_tuple_list": cc_field_tuple_list("field_name"),
        "c_field_id_tuple_list": cc_field_id_tuple_list("field_name"),
        "c_field_not_none_tuple_list": cc_field_not_none_tuple_list(
            "field_name"
        ),
        "c_field_id_not_none_tuple_list": (
            cc_field_id_not_none_tuple_list("field_name")
        ),
        "c_field_none_tuple_list": cc_field_none_tuple_list(
            "field_name"
        ),
        "c_field_id_none_tuple_list": cc_field_id_none_tuple_list(
            "field_name"
        ),
        "c_field_tuple_empty": cc_field_tuple_empty("field_name"),
        "c_field_id_tuple_empty": cc_field_id_tuple_empty(
            "field_name"
        ),
        "c_field_not_none_tuple_empty": cc_field_not_none_tuple_empty(
            "field_name"
        ),
        "c_field_id_not_none_tuple_empty": (
            cc_field_id_not_none_tuple_empty("field_name")
        ),
        "c_field_none_tuple_empty": cc_field_none_tuple_empty(
            "field_name"
        ),
        "c_field_id_none_tuple_empty": cc_field_id_none_tuple_empty(
            "field_name"
        ),
        "c_field_list_empty": cc_field_list_empty("field_name"),
        "c_field_id_list_empty": cc_field_id_list_empty("field_name"),
        "c_field_not_none_list_empty": cc_field_not_none_list_empty(
            "field_name"
        ),
        "c_field_id_not_none_list_empty": (
            cc_field_id_not_none_list_empty("field_name")
        ),
        "c_field_none_list_empty": cc_field_none_list_empty(
            "field_name"
        ),
        "c_field_id_none_list_empty": cc_field_id_none_list_empty(
            "field_name"
        ),
        "c_field_list_list_nnl": cc_field_list_list_nnl("field_name"),
        "c_field_id_list_list_nnl": cc_field_id_list_list_nnl(
            "field_name"
        ),
        "c_field_not_none_list_list_nnl": (
            cc_field_not_none_list_list_nnl("field_name")
        ),
        "c_field_id_not_none_list_list_nnl": (
            cc_field_id_not_none_list_list_nnl("field_name")
        ),
        "c_field_none_list_list_nnl": cc_field_none_list_list_nnl(
            "field_name"
        ),
        "c_field_id_none_list_list_nnl": (
            cc_field_id_none_list_list_nnl("field_name")
        ),
        "c_field_tuple_tuple_nnl": cc_field_tuple_tuple_nnl(
            "field_name"
        ),
        "c_field_id_tuple_tuple_nnl": cc_field_id_tuple_tuple_nnl(
            "field_name"
        ),
        "c_field_not_none_tuple_tuple_nnl": (
            cc_field_not_none_tuple_tuple_nnl("field_name")
        ),
        "c_field_id_not_none_tuple_tuple_nnl": (
            cc_field_id_not_none_tuple_tuple_nnl("field_name")
        ),
        "c_field_none_tuple_tuple_nnl": (
            cc_field_none_tuple_tuple_nnl("field_name")
        ),
        "c_field_id_none_tuple_tuple_nnl": (
            cc_field_id_none_tuple_tuple_nnl("field_name")
        ),
        "c_field_list_tuple_nnl": (
            cc_field_list_tuple_nnl("field_name")
        ),
        "c_field_id_list_tuple_nnl": cc_field_id_list_tuple_nnl(
            "field_name"
        ),
        "c_field_not_none_list_tuple_nnl": (
            cc_field_not_none_list_tuple_nnl("field_name")
        ),
        "c_field_id_not_none_list_tuple_nnl": (
            cc_field_id_not_none_list_tuple_nnl("field_name")
        ),
        "c_field_none_list_tuple_nnl": cc_field_none_list_tuple_nnl(
            "field_name"
        ),
        "c_field_id_none_list_tuple_nnl": (
            cc_field_id_none_list_tuple_nnl("field_name")
        ),
        "c_field_tuple_list_nnl": cc_field_tuple_list_nnl(
            "field_name"
        ),
        "c_field_id_tuple_list_nnl": cc_field_id_tuple_list_nnl(
            "field_name"
        ),
        "c_field_not_none_tuple_list_nnl": (
            cc_field_not_none_tuple_list_nnl("field_name")
        ),
        "c_field_id_not_none_tuple_list_nnl": (
            cc_field_id_not_none_tuple_list_nnl("field_name")
        ),
        "c_field_none_tuple_list_nnl": cc_field_none_tuple_list_nnl(
            "field_name"
        ),
        "c_field_id_none_tuple_list_nnl": (
            cc_field_id_none_tuple_list_nnl("field_name")
        ),
        "c_field_tuple_empty_nnl": cc_field_tuple_empty_nnl(
            "field_name"
        ),
        "c_field_id_tuple_empty_nnl": cc_field_id_tuple_empty_nnl(
            "field_name"
        ),
        "c_field_not_none_tuple_empty_nnl": (
            cc_field_not_none_tuple_empty_nnl("field_name")
        ),
        "c_field_id_not_none_tuple_empty_nnl": (
            cc_field_id_not_none_tuple_empty_nnl("field_name")
        ),
        "c_field_none_tuple_empty_nnl": cc_field_none_tuple_empty_nnl(
            "field_name"
        ),
        "c_field_id_none_tuple_empty_nnl": (
            cc_field_id_none_tuple_empty_nnl("field_name")
        ),
        "c_field_list_empty_nnl": cc_field_list_empty_nnl(
            "field_name"
        ),
        "c_field_id_list_empty_nnl": cc_field_id_list_empty_nnl(
            "field_name"
        ),
        "c_field_not_none_list_empty_nnl": (
            cc_field_not_none_list_empty_nnl("field_name")
        ),
        "c_field_id_not_none_list_empty_nnl": (
            cc_field_id_not_none_list_empty_nnl("field_name")
        ),
        "c_field_none_list_empty_nnl": cc_field_none_list_empty_nnl(
            "field_name"
        ),
        "c_field_id_none_list_empty_nnl": (
            cc_field_id_none_list_empty_nnl("field_name")
        ),
    }
)

NUMBER_OF_ITERATIONS = 1000000


def benchmark(callbacks_to_benchmark):
    """
    Benchmark all the callbacks in the argument dict.
    """

    result = []
    for name, callback in callbacks_to_benchmark.items():
        time_with_fk = timeit.timeit(
            functools.partial(callback, instance_with_fk),
            number=NUMBER_OF_ITERATIONS,
        )
        time_without_fk = timeit.timeit(
            functools.partial(callback, instance_without_fk),
            number=NUMBER_OF_ITERATIONS,
        )
        result.append(
            {
                "name": name,
                "time_with_fk": time_with_fk,
                "time_without_fk": time_without_fk,
            }
        )
    result.sort(key=lambda x: x["time_without_fk"])
    result.sort(key=lambda x: x["time_with_fk"])
    for data in result:
        print(
            f"{data['name']}:"
            f" Avec FK: {data['time_with_fk']},"
            f" Sans FK: {data['time_without_fk']}"
        )


# Benchmark each callback.
benchmark(callbacks)

print("-------------------------------------------------------------")

# Benchmark each callback in random order.
keys = list(callbacks.keys())
random.shuffle(keys)
callbacks_2 = {key: callbacks[key] for key in keys}
benchmark(callbacks_2)



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
In older versions of Django, a prefetch use the ordering defined
on the class prefetched.
class Meta:
    ordering = ...
But when you prefetch and there is a single value per current object
that will recieve the prefetch in its cache,
then the ordering is totally useless.
It adds useless load on the DB server.
The goal of these patches is to remove this order by
for many versions of Django.
On very big prefetches I noticed a gain of 10 - 15 %,
since the rest of the DB query is mainly an index scan.
#ClimateChangeBrake
Note that I also submitted a ticket and a PR:
https://code.djangoproject.com/ticket/35309
https://github.com/django/django/pull/17984
and it was merged, hence Django 5.1 and afterward
will not have this problem,
and maybe older versions if it is backported to some.
"""

# pylint: disable-next=import-error
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)

# In old versions of Django,
# the method is called get_prefetch_queryset().
# It is deprecated in Django 5 and will be removed in Django 6.
# Apply the patch that corresponds to your version.
# pylint: disable-next=invalid-name
original_forward_many_to_one_get_prefetch_queryset = None
if hasattr(ForwardManyToOneDescriptor, "get_prefetch_queryset"):
    original_forward_many_to_one_get_prefetch_queryset = (
        ForwardManyToOneDescriptor.get_prefetch_queryset
    )
# pylint: disable-next=invalid-name
original_reverse_one_to_one_get_prefetch_queryset = None
if hasattr(ReverseOneToOneDescriptor, "get_prefetch_queryset"):
    original_reverse_one_to_one_get_prefetch_queryset = (
        ReverseOneToOneDescriptor.get_prefetch_queryset
    )

# pylint: disable-next=invalid-name
original_forward_many_to_one_get_prefetch_querysets = None
if hasattr(ForwardManyToOneDescriptor, "get_prefetch_querysets"):
    original_forward_many_to_one_get_prefetch_querysets = (
        ForwardManyToOneDescriptor.get_prefetch_querysets
    )
# pylint: disable-next=invalid-name
original_reverse_one_to_one_get_prefetch_querysets = None
if hasattr(ReverseOneToOneDescriptor, "get_prefetch_querysets"):
    original_reverse_one_to_one_get_prefetch_querysets = (
        ReverseOneToOneDescriptor.get_prefetch_querysets
    )


def get_correct_queryset(self, queryset):
    """
    The patched way to init the queryset in get_prefetch_queryset().
    """
    if queryset is None:
        queryset = self.get_queryset()
    return queryset.order_by()


def patched_forward_many_to_one_get_prefetch_queryset_v1(
    self, instances, queryset=None
):
    """
    The patch for ForwardManyToOneDescriptor.get_prefetch_queryset().
    """
    queryset = get_correct_queryset(self, queryset)
    return original_forward_many_to_one_get_prefetch_queryset(
        self, instances, queryset
    )


def patched_reverse_one_to_one_get_prefetch_queryset_v1(
    self, instances, queryset=None
):
    """
    The patch for ReverseOneToOneDescriptor.get_prefetch_queryset().
    """
    queryset = get_correct_queryset(self, queryset)
    return original_reverse_one_to_one_get_prefetch_queryset(
        self, instances, queryset
    )


def apply_patched_get_prefetch_queryset_v1():
    """
    Applying the patch for versions of Django with
    get_prefetch_queryset().
    """
    ForwardManyToOneDescriptor.get_prefetch_queryset = (
        patched_forward_many_to_one_get_prefetch_queryset_v1
    )
    ReverseOneToOneDescriptor.get_prefetch_queryset = (
        patched_reverse_one_to_one_get_prefetch_queryset_v1
    )


def get_correct_querysets(self, querysets):
    """
    The patched way to init the querysets in get_prefetch_querysets().
    """
    if querysets is None:
        querysets = [self.get_queryset()]
    if len(querysets) == 1:
        querysets[0] = querysets[0].order_by()
    return querysets


def patched_forward_many_to_one_get_prefetch_querysets_v1(
    self, instances, querysets=None
):
    """
    The patch for ForwardManyToOneDescriptor.get_prefetch_querysets().
    """
    querysets = get_correct_querysets(self, querysets)
    return original_forward_many_to_one_get_prefetch_querysets(
        self, instances, querysets
    )


def patched_reverse_one_to_one_get_prefetch_querysets_v1(
    self, instances, querysets=None
):
    """
    The patch for ReverseOneToOneDescriptor.get_prefetch_querysets().
    """
    querysets = get_correct_querysets(self, querysets)
    return original_reverse_one_to_one_get_prefetch_querysets(
        self, instances, querysets
    )


def apply_patched_get_prefetch_querysets_v1():
    """
    Applying the patch for versions of Django with
    get_prefetch_querysets().
    """
    ForwardManyToOneDescriptor.get_prefetch_querysets = (
        patched_forward_many_to_one_get_prefetch_querysets_v1
    )
    ReverseOneToOneDescriptor.get_prefetch_querysets = (
        patched_reverse_one_to_one_get_prefetch_querysets_v1
    )



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
To optimize some queries, you may need to add prefetches
that apply only to some of the instances.
It is very convenient when you have some expensive prefetches
that concerns only a few instances.
Hence, I added a filter_callback argument to Prefetch.
#ClimateChangeBrake
Note that I also submitted a ticket and a PR:
https://code.djangoproject.com/ticket/35317
https://github.com/django/django/pull/17994
Unfortunately, it was rejected without really considering
the true value of it.
Moreover, I had very special use cases that needed also to add
a post_prefetch_callback argument to Prefetch.
Below is the code enabling all this.
You may not see the need for such things at first.
But if you have some time, here is a redacted real world example
of its use; at work, we use Django Rest Framework,
with lots of nested serializers.
In each model serializer, I add a static method enhance_queryset()
calling the methods enhance_queryset() of the nested serializers, etc.
So compare toy examples with production work...


class OModelSerializer(...):
    @staticmethod
    def enhance_queryset(
        query_set,
        prefix: str = "",
        c_id: Optional[CId] = None,
    ):
        query_set = SomeMixin1.enhance_queryset(query_set, prefix)

        def ventilate_ps_by_c_id(
            lookup,
            done_queries,
        ):
            prefetch_to = lookup.prefetch_to
            prefetch_before = get_previous_prefetch_to(prefetch_to)
            for obj in prefetch_before:
                if hasattr(obj, "needed_ps"):
                    if not hasattr(obj, "needed_p_by"):
                        obj.needed_p_by = {}
                    for obj2 in obj.needed_ps:
                        obj.needed_p_by[obj2.c_id] = obj2

        return query_set.prefetch_related(
            f"{prefix}p2",
            Prefetch(
                f"{prefix}s",
                post_prefetch_callback=(
                  create_post_prefetch_callback_add_backward_multiple(
                    retrieve_forward_cache_callback=lambda o: [o.s]
                    if o.s_id
                    else [],
                    backward_cache_name="current_o_ancestors",
                  )
                ),
            ),
            Prefetch(
                f"{prefix}c2",
                post_prefetch_callback=(
                  create_post_prefetch_callback_add_backward_multiple(
                    retrieve_forward_cache_callback=lambda o:[o.c2]
                    if o.c2_id
                    else [],
                    backward_cache_name="current_order_ancestors",
                  )
                ),
            ),
            Prefetch(
                f"{prefix}s__ps",
                queryset=(
                    P.objects.filter(c_id=c_id)
                    if c_id
                    else P.objects.all()
                ),
                to_attr="needed_ps",
                filter_callback=(
                  lambda p: hasattr(p, "_prefetched_objects_cache")
                )
                and p._prefetched_objects_cache.get(
                    "current_order_ancestors"
                )
                and any(
                    map(
                        lambda o: o.c2_id is not None,
                        p._prefetched_objects_cache.get(
                            "current_o_ancestors"
                        ).values(),
                    )
                ),
                post_prefetch_callback=ventilate_ps_by_c_id,
            ),
            Prefetch(
                f"{prefix}c2__u",
                queryset=C2U.objects.filter(p2_id__isnull=False)
                .distinct("c2_id")
                .prefetch_related(
                    "p2",
                    Prefetch(
                        f"p2__ps",
                        queryset=(
                            P.objects.filter(c_id=c_id)
                            if c_id
                            else P.objects.all()
                        ),
                        to_attr="needed_ps",
                        post_prefetch_callback=ventilate_ps_by_c_id,
                    ),
                ),
                to_attr="needed_u",
                filter_callback=lambda c: (
                    hasattr(c2, "_prefetched_objects_cache")
                    and c2._prefetched_objects_cache.get(
                        "current_o_ancestors"
                    )
                    and any(
                        map(
                            lambda o: o.c2_id is not None
                            and o.s_id is None,
                            c._prefetched_objects_cache.get(
                                "current_o_ancestors"
                            ).values(),
                        )
                    )
                ),
            ),
        )


class DModelSerializer(...):
    ...

    @staticmethod
    def enhance_queryset(
        query_set,
        prefix: str = "",
        c_id: Optional[CId] = None,
    ):
        query_set = SomeMixin2.enhance_queryset(
            query_set,
            f"{prefix}l__",
        )
        query_set = OModelSerializer.enhance_queryset(
            query_set,
            f"{prefix}l__",
            c_id=c_id,
        )
        query_set = query_set.prefetch_related(
            f"{prefix}l__s2",
            f"{prefix}l__p3",
            f"{prefix}l2__s3__u",
            Prefetch(
                f"{prefix}l__r1",
                queryset=R1.objects.filter(
                    d_id=F("o__s2__d_id"),
                    r2__s4=SOME_CONSTANT,
                ).select_related("r2"),
                to_attr="pertinent_r3",
            ),
        )

        return query_set

So maybe now, you understand why
I do not understand framework maintainers
that block any new feature not coming from them.
I think we don't live in code of the same complexity.
I did a second PR,
mainly because it can help (but it was rejected as usual),
and because it can help for understanding this patch,
and see that the modifications needed are only minor:
https://github.com/django/django/pull/18003
"""

import importlib

# pylint: disable-next=import-error
from django.db.models import Prefetch, query

# See exec below for the need of these imports.
# pylint: disable-next=import-error,unused-import
from django.db.models.query import (
    LOOKUP_SEP,
    get_prefetcher,
    normalize_prefetch_lookups,
)

old_prefetch__init__ = Prefetch.__init__


def patched_prefetch__init__v1(
    self, lookup, queryset=None, to_attr=None, filter_callback=None
):
    """
    Patch for Prefetch.__init__ with filter_callback kwarg.
    """
    old_prefetch__init__(
        self, lookup, queryset=queryset, to_attr=to_attr
    )
    self.filter_callback = filter_callback


# pylint: disable-next=too-many-arguments
def patched_prefetch__init__v2(
    self,
    lookup,
    queryset=None,
    to_attr=None,
    filter_callback=None,
    post_prefetch_callback=None,
):
    """
    Patch for Prefetch.__init__ with filter_callback
    and post_prefetch_callback kwargs.
    """
    old_prefetch__init__(
        self, lookup, queryset=queryset, to_attr=to_attr
    )
    self.filter_callback = filter_callback
    self.post_prefetch_callback = post_prefetch_callback


old_prefetch_one_level = query.prefetch_one_level


def patched_prefetch_one_level_v1(
    instances, prefetcher, lookup, level
):
    """
    Patch for prefetch_one_level()
    applying Prefetch.filter_callback.
    """
    if lookup.filter_callback is not None:
        instances = [
            instance
            for instance in instances
            if lookup.filter_callback(instance)
        ]
    if len(instances) == 0:
        return [], ()
    return old_prefetch_one_level(
        instances, prefetcher, lookup, level
    )


def apply_patch_prefetch_with_filter_callback_v1():
    """
    Applying the patch for prefetches on subsets of instances.
    """
    Prefetch.__init__ = patched_prefetch__init__v1
    query.prefetch_one_level = patched_prefetch_one_level_v1


# Now things are starting to be amusing.
# We load the source code of the function prefetch_related_objects()
# in the installed version of Django and we patch its source code
# with string manipulations.
# First time for me in Python (not in PHP). :)
# It works with Django 2.2 and Django 5.0,
# probably also with versions in-between and after.
spec = importlib.util.find_spec("django.db.models.query", "django")
source = spec.loader.get_source("django.db.models.query")
source = source[
    source.find("def prefetch_related_objects") : source.find(
        "obj_list = new_obj_list"
    )
    + len("obj_list = new_obj_list")
]
source = (
    source.replace(
        "def prefetch_related_objects",
        "def patched_prefetch_related_objects_v1",
    )
    .replace(
        "done_queries = {}",
        'done_queries = {"": model_instances}',
    )
    .replace(
        "if lookup.prefetch_to in done_queries:",
        "if lookup.prefetch_to in done_queries:\n"
        "            if lookup.post_prefetch_callback:\n"
        "                lookup.post_prefetch_callback("
        "lookup, done_queries)\n",
    )
    .replace(
        "obj_list = new_obj_list",
        "obj_list = new_obj_list\n"
        "        if lookup.post_prefetch_callback:\n"
        "            lookup.post_prefetch_callback("
        "lookup, done_queries)\n",
    )
)


# local value needed for the exec
prefetch_one_level = patched_prefetch_one_level_v1

# This line is not needed;
# in fact exec below will define the variable.
# But it is clearer.
# pylint: disable-next=invalid-name
patched_prefetch_related_objects_v1 = None

# pylint: disable-next=exec-used
exec(
    source,
    globals(),
    locals(),
)


def apply_patch_prefetch_with_v2():
    """
    Applying the patch for prefetches with:
    - filter_callback,
    - post_prefetch_callback.
    """
    Prefetch.__init__ = patched_prefetch__init__v2
    query.prefetch_one_level = patched_prefetch_one_level_v1
    query.prefetch_related_objects = (
        patched_prefetch_related_objects_v1
    )


def get_previous_prefetch_to(prefetch_to):
    """
    Going one step before in the order of prefetches.
    """
    if prefetch_to == "":
        raise ValueError(
            "get_previous_prefetch_to()"
            " You're going backward too much."
        )
    prefetch_before = ""
    index = prefetch_to.rfind(LOOKUP_SEP)
    if index != -1:
        prefetch_before = prefetch_to[:index]
    return prefetch_before


def create_post_prefetch_callback_add_backward_multiple(
    # Examples:
    # retrieve_forward_cache_callback=
    #   lambda x: [x.my_foreign_key] if x.my_foreign_key else []
    # retrieve_forward_cache_callback=
    #   lambda x: x._prefetched_objects_cache.get("toto", [])
    retrieve_forward_cache_callback,
    # backward_cache will always go
    # in x._prefetched_objects_cache as a dict
    # with a unique key to avoid repeating and the object.
    backward_cache_name,
    backward_level=0,
    # defaults to obj.pk
    get_key_for_backward_object_callback=None,
):
    """
    A function to obtain post_prefetch_callbacks
    filling backward relationships when multiple objects
    can prefetch the same object.
    Note that when the models relationships allow only
    one single object to prefetch the same object,
    then the framework already fills the backward cache.
    """

    def post_prefetch_callback_add_backward_multiple(
        lookup,
        done_queries,
    ):
        """
        The customized post_prefetch_callback function
        that will be returned.
        """
        nonlocal retrieve_forward_cache_callback
        nonlocal backward_cache_name
        nonlocal backward_level
        nonlocal get_key_for_backward_object_callback

        prefetch_to = lookup.prefetch_to
        while backward_level > 0:
            prefetch_to = get_previous_prefetch_to(prefetch_to)
            backward_level -= 1
        prefetch_before = get_previous_prefetch_to(prefetch_to)
        for obj in done_queries[prefetch_before]:
            forward_objects = retrieve_forward_cache_callback(obj)
            if get_key_for_backward_object_callback is None:
                key = obj.pk
            else:
                key = get_key_for_backward_object_callback(obj)
            for obj2 in forward_objects:
                if not hasattr(obj2, "_prefetched_objects_cache"):
                    # pylint: disable-next=protected-access
                    obj2._prefetched_objects_cache = {}
                if (
                    # pylint: disable-next=protected-access
                    obj2._prefetched_objects_cache.get(
                        backward_cache_name
                    )
                    is None
                ):
                    # pylint: disable-next=protected-access
                    obj2._prefetched_objects_cache[
                        backward_cache_name
                    ] = {}
                # pylint: disable-next=protected-access
                obj2._prefetched_objects_cache[backward_cache_name][
                    key
                ] = obj

    return post_prefetch_callback_add_backward_multiple


e = ()  # empty_tuple


def create_retrieve_forward_cache_callback_for_foreign_key(
    f,  # field_name
):
    """
    A function to obtain a retrieve_forward_cache_callback
    corresponding to a foreign key of a Django model.
    """
    g = f + "_id"

    def retrieve_forward_cache_callback_for_foreign_key(x):
        """
        The customized retrieve_forward_cache_callback function
        that will be returned.
        """

        return e if getattr(x, g) is None else (getattr(x, f),)

    return retrieve_forward_cache_callback_for_foreign_key


def create_retrieve_forward_cache_callback_for_prefetched_objects(
    c,  # cache_name
):
    """
    A function to obtain a retrieve_forward_cache_callback
    corresponding to an entry in _prefetched_objects_cache
    of a Django model.
    """

    def retrieve_forward_cache_callback_for_prefetched_objects(x):
        """
        The customized retrieve_forward_cache_callback function
        that will be returned.
        """

        if not hasattr(x, "_prefetched_objects_cache"):
            return e

        # pylint: disable-next=protected-access
        return x._prefetched_objects_cache.get(c, e)

    return retrieve_forward_cache_callback_for_prefetched_objects



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
A collection of functions to make the key used by the cache.
As of now, it is only for tests.
"""

import os


def make_test_cache_key(key, key_prefix, version):
    """
    Add a simple "test_cache" prefix.

    For example, apply it in your settings file with:
    # if len(sys.argv) > 1 and sys.argv[1] == "test":
    if TESTING:
        for cache in CACHES.values():
            cache["KEY_FUNCTION"] = make_test_cache_key
    """
    return f"test_cache:{key_prefix}:{version}:{key}"


def make_parallel_test_cache_key(key, key_prefix, version):
    """
    Add a simple "test_cache" prefix plus the current pid.
    The local memory cache is isolated between processes.
    But it may not be the case of a Redis backend cache, for example.
    https://stackoverflow.com/questions/71583690/
    It may seem inefficient to call os.getpid() here,
    but Django loads settings only once even when multiple processes
    are used for parallel tests.
    And the BaseCache related objects are common.
    So setting KEY_PREFIX is not enough.

    For example, apply it in your settings file with:
    # if len(sys.argv) > 1 and sys.argv[1] == "test":
    if TESTING:
        for cache in CACHES.values():
            cache["KEY_FUNCTION"] = make_parallel_test_cache_key
    """

    return f"test_cache{os.getpid()}:{key_prefix}:{version}:{key}"



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
A full-featured custom query wrapper and constants to control it,
plus functions to customize part of its behavior.
The goal is to give back full control to the developper
by enabling debug, etc. that will be localized to a problem to solve
without making the whole application unusable/slow in production.
"""

import re
import time
import traceback

# pylint: disable-next=import-error
from django.db import connections

# Remind that you can modify the constants below during execution:
# from django_monkey_patches import django__query_wrapper
# django__query_wrapper.COUNT_QUERIES = True
# ... the code I'm looking at closely
# django__query_wrapper.COUNT_QUERIES = False
# logger.info(f"Number of queries : {...}")

# Why you may want to throttle queries?
# To see the impact of some conditions,
# like when you have an n-tier architecture
# and you always have a penalty of approximately 3 ms per query.
# You can somehow reproduce this in your development environment
# to see where you may have a problem later.
THROTTLE_QUERIES = False
QUERY_PENALTY_MILLISECONDS = 3
# When the number of queries "scales" too closely
# with the number of objects processed...
# The count will be saved on connection(s) object(s).
COUNT_QUERIES = False
# When you have some particular queries that are too slow.
# If you can control settings.DEBUG,
# you may want to use a log filter on django.db.backends.
# But settings.DEBUG activates too many things
# (bad frameworks use global constants that do everything,
# good frameworks have atomic constants for anything to work
# in isolation,
# better frameworks choose carefully
# the intermediate switches to add to the good frameworks.)
# Bizarre, il me semblait déjà avoir écrit ça en commençant
# à faire ce patch en novembre ou décembre 2023 ?
# Mais là, plus de trace sur mon ordi.
# Je me fais laver le cerveau un peu trop souvent.
# Autre truc marrant, la recherche GitHub sur le mot atomic
# ne trouve rien dans mon repo django-monkey-patches,
# alors que le push a plus de 5 heures d'ancienneté
# (grep c'est compliqué comme feature ?).
TIME_QUERIES = False
# Activate the dump of the call stack to process or log it.
COMPUTE_CALL_STACK = False
# Execute a callback before executing the query on the DBMS.
PRE_EXECUTION_CALLBACK = None
# Execute a callback after executing the query on the DBMS.
POST_EXECUTION_CALLBACK = None

# All functions below are suffixed with _v1.
# Beware, that I may not use the strictest interpretation
# of backward compatible, because it would need to duplicate
# hundreds of lines of code each time.
# If you understand the logic, you will see that
# the backward compatible problems I don't want to handle
# is about adding custom "fields" in the data structure.
# Please, look at the message at the beginning of
# get_extra_data_template_for_set_of_queries_v1(),
# and submit an issue if needed.
# Otherwise, there may be a "new v1" that will break your code.
# The other backward compatible problems should normally be adressed.


def custom_query_wrapper_v1(execute, sql, params, many, context):
    """
    This custom query wrapper features many options
    to help debug and profile database queries.
    You should be able to do anything you want with it.
    Below, are given data structures and a POST_EXECUTION_CALLBACK
    that should handle most of your use cases.

    Here is an example of how to use it with a middleware:

    class CustomQueryWrapperMiddleware:
        def __init__(self, get_response):
            self.get_response = get_response

        def __call__(self, request):
            # Optional but recommended, see below.
            # At least an equivalent is needed,
            # if you activate COUNT_QUERIES.
            init_connections_extra_data_v1()

            # See at end of file to wrap all connections.
            with connection.execute_wrapper(custom_query_wrapper_v1):
                return self.get_response(request)

    In settings.py, add:
    # ######### MIDDLEWARE CONFIGURATION
    MIDDLEWARE = (
        ...
        "path_to.middlewares.CustomQueryWrapperMiddleware",
        ...
    )

    If you want to use it in your commands, there is more work.
    I recommend to define your own class:
    class CustomBaseCommand(BaseCommand):
        ...
        def execute(self, *args, **options):
            # Copy the code of BaseCommand.execute()
            # And in it replace at least:
            #   output = self.handle(*args, **options)
            # with:
            with connection.execute_wrapper(custom_query_wrapper_v1):
                # See at end of file to wrap all connections.
                output = self.handle(*args, **options)
        ...
    And then, make all your commands inherit from CustomBaseCommand.
    You can go way further
    in this line of customization and monitoring.
    """

    if THROTTLE_QUERIES:
        time.sleep(QUERY_PENALTY_MILLISECONDS / 1000)

    call_stack = None
    if COMPUTE_CALL_STACK:
        call_stack = traceback.format_stack()

    if PRE_EXECUTION_CALLBACK is not None:
        (
            execute,
            sql,
            params,
            many,
            context,
            call_stack,
        ) = PRE_EXECUTION_CALLBACK(
            execute, sql, params, many, context, call_stack
        )

    start_time = None
    if TIME_QUERIES:
        start_time = time.time()

    result = execute(sql, params, many, context)

    end_time = None
    duration = None
    if TIME_QUERIES:
        end_time = time.time()
        duration = end_time - start_time

    if COUNT_QUERIES:
        connections.django_monkey_patches_dict["query_count"] += 1
        connection = context["connection"]
        connection.django_monkey_patches_dict["query_count"] += 1

    if POST_EXECUTION_CALLBACK is not None:
        result = POST_EXECUTION_CALLBACK(
            execute,
            sql,
            params,
            many,
            context,
            call_stack,
            start_time,
            result,
            end_time,
            duration,
        )

    return result


# pylint: disable-next=unused-argument
def get_full_query_v1(extra_data_dict, data):
    """
    An helper function to see the request
    as it will be transmitted to the DBMS.
    """

    connection = data["context"]["connection"]
    return connection.cursor().mogrify(data["sql"], data["params"])


get_full_query_v1.field_name = "full_query_v1"


params_placeholders_regexp = re.compile("%s(,%s)*")


# pylint: disable-next=unused-argument
def get_sql_signature_v1(extra_data_dict, data):
    """
    An helper function to get an SQL signature
    where the params have no impact.
    It is somewhat similar to a technique we used at Teliae.
    Jérôme Petit added an activable dump of all queries
    and progressively improved the profiling tool,
    with help of other team members including me,
    this kind of signature was used to group requests together.
    But since our in-house framework was rarely using
    prepared statements,
    the code to obtain an "SQL signature" was different.
    All the code in this file is more general and flexible than
    what we had at Teliae when I was working there.
    But the ideas here both stem from my work at Teliae and Deleev.
    """

    return params_placeholders_regexp.sub("", data["sql"])


get_sql_signature_v1.field_name = "sql_signature_v1"


# pylint: disable-next=unused-argument
def get_light_call_stack_v1(extra_data_dict, data):
    """
    An helper function to remove framework code from the call stack.
    Usually, you want to know where you generate a DB request
    in your own code.
    """

    return [
        line
        for line in data["call_stack"]
        if "site-packages/" not in line
    ]


get_light_call_stack_v1.field_name = "light_call_stack_v1"


def get_managed_list(manager):
    """
    Smaller with auto-completion:
    get_managed_list(manager)
    [] if manager is None else manager.list()
    """

    if manager is None:
        return []
    return manager.list()


def get_managed_dict(manager):
    """
    Smaller with auto-completion:
    get_managed_dict(manager)
    {} if manager is None else manager.dict()
    """

    if manager is None:
        return {}
    return manager.dict()


# Unfortunately, there is no "inline" function or macro in Python.
# Another strange fact, C has inline functions,
# but no keyword to inline only at some spot.
# And sometimes, you know where is the spot
# where it MUST-epsilon be inlined.
# In theory, the compiler is supposed to know better than you...
# Search for examples.
# Of course, you can add assembly in C code instead.
# Much more convenient.
# Or you can use ORDER or CHAOS C "macros engines",
# and recently Zig, but I would prefer more keywords
# and less limited base macro rules for C,
# instead of learning something totally new.
# There may be more advantages to Zig that I don't see,
# since I only read news about it.
# And there is Rust with partial memory-safety.
# But since I almost never do multi-threaded code,
# I'm more looking at C and languages enabling very fast code,
# than more safe code.
# https://github.com/google/comprehensive-rust/issues
# Something disappeared : I have no "issue",
# but I still have some memory and a brain and a shiny armor
# and a sword in the hands of God:
# Let's hope it is just my tongue
# to tell the Truth in the name of Jesus.
# Let's hope I shall use my tongue more frequently aligned/adjusted
# with God's intention. I have no pretention of perfection.
# fn main() {
#   for x in 1..5 {
#     println!("x: {x}");
#   }
#   let bla = [1, 2, 3, 4, 5];
#   // La bonne excuse,
#   // c'est l'indécidabilité du problème de l'arrêt
#   // (cf. le film Didier, certains ne savent pas s'arrêter),
#   // style Collatz conjecture,
#   // c'est clair qu'on fait ce genre de blagues
#   // dans le code tout le temps...
#   // Non mais sérieux... 99 % du code,
#   // il doit y avoir des règles simples
#   // pour ne pas avoir à gérer de problème indécidable...
#   // Enfin, ce n'est que mon avis.
#   // Mais je ne bosse pas trop sur des compilateurs ;
#   // peut-être que c'est un peu plus au niveau méta
#   // qu'il y a des trucs "amusants" sur le plan scientifique,
#   // et encore ce n'est pas une certitude.
#
#   for elem in 1..10 {
#     let truc = bla[elem];
#     println!("elem: {truc}");
#   }
# }


# pylint: disable-next=too-many-arguments
def get_extra_data_template_for_set_of_queries_v1(
    query_fields=None,
    min_seconds_threshold=0,
    max_seconds_threshold=float("inf"),
    filter_callback=None,
    insertion_callback=None,
    subsets_extra_data=None,
    allocated_subsets_extra_data=None,
    allocated_subsets_key_callback=None,
    allocated_subsets_init_callback=None,
    top_down_post_processing_callback=None,
    bottom_up_post_processing_callback=None,
    # You may need that with django-rq or other multiprocesses code:
    manager=None,
):
    """
    Obtain a default extra_data_dict with most of
    the interesting "fields" regarding DB queries analysis.
    If you see something that may be interesting to add here,
    with a good example of why you need it, please,
    submit an issue on my GitHub repository.
    That's the goal of the LGPL: if you improve this library code,
    please share with anyone.
    I will not sue you if you don't, since it is a border-line case.
    Just adding "fields" and then using other functions
    without my code
    or eventually wrapping mines may be subject to interpretation.
    Basically, I think that if you use your own additional "fields"
    and your own callbacks to fill/use the "fields" below,
    you have no obligation at all.
    But it would be nice to contribute anyway.
    """

    if query_fields is None:
        query_fields = get_managed_list(manager)
    if subsets_extra_data is None:
        subsets_extra_data = get_managed_dict(manager)
    if allocated_subsets_extra_data is None:
        allocated_subsets_extra_data = get_managed_dict(manager)
    if allocated_subsets_key_callback is None:
        allocated_subsets_key_callback = get_managed_dict(manager)
    if allocated_subsets_init_callback is None:
        allocated_subsets_init_callback = get_managed_dict(manager)

    result = get_managed_dict(manager)
    result_content = {
        "query_count": 0,
        "query_fields": query_fields,  # [
        #     "execute",
        #     "sql",
        #     "params",
        #     "many",
        #     "context",
        #     "call_stack",
        #     "start_time",
        #     "result",
        #     "end_time",
        #     "duration",
        #     get_full_query_v1,  # This is a function
        # ],
        "query_list": get_managed_list(manager),  # [
        # {
        #     "execute": None,
        #     "sql": None,
        #     "params": None,
        #     "many": None,
        #     "context": None,
        #     "call_stack": None,
        #     "start_time": None,
        #     "result": None,
        #     "end_time": None,
        #     "duration": None,
        #     "full_query_v1": None,
        # },
        # ],
        "total_duration": 0,
        "average_duration": 0,
        "min_duration": float("inf"),
        "max_duration": 0,
        # ------------------------------------------------------------
        # These fields could be at the top,
        # but it would be less didactic I think.
        # Two floats that will be compared to query duration
        # for any query in ancestor subset;
        # in case a comparison fails,
        # the query is not counted in this subset.
        "min_seconds_threshold": min_seconds_threshold,  # 0,
        # float("inf"),
        "max_seconds_threshold": max_seconds_threshold,
        "filter_callback": filter_callback,  # None,
        # Default will be to use query_fields, update total,
        # min, max, and recurse on nested dicts.
        "insertion_callback": insertion_callback,  # None,
        # ------------------------------------------------------------
        "subsets_extra_data": subsets_extra_data,  # {
        #     "main": get_extra_data_template_for_set_of_queries(),
        # },
        "allocated_subsets_extra_data": allocated_subsets_extra_data,
        # {
        #   Beware of recursive calls
        #   https://stackoverflow.com/questions/
        #   22464900/do-dictionaries-have-a-key-length-limit
        #   Value of dict can be the count or a full nested dict.
        #   "distinct_call_stacks": {
        #     f"{some_call_stack}": (
        #       get_extra_data_template_for_set_of_queries(),
        #     )
        #   },
        # },
        "allocated_subsets_key_callback": (
            allocated_subsets_key_callback
            # {
            #   "distinct_call_stacks": lambda x, y: y["call_stack"],
            # },
        ),
        "allocated_subsets_init_callback": (
            allocated_subsets_init_callback
            # {
            #   "distinct_call_stacks": (
            #     get_extra_data_template_for_set_of_queries
            #   ),
            # },
        ),
        # If you want an allocated_subsets_filter_callback,
        # because you want to allocate only a subset of queries,
        # then you're invited to use one more nesting level
        # with subsets_extra_data.
        # Because it will avoid duplicating
        # the, currently, three fields
        # used for filtering above.
        # With the following callbacks, you can sort query_list,
        # for example.
        # But you may also reorder the nested dicts, etc.,
        # and you can control if you want to synthetize
        # bottom-up or top_down.
        "top_down_post_processing_callback": (
            top_down_post_processing_callback  # None
        ),
        "bottom_up_post_processing_callback": (
            bottom_up_post_processing_callback  # None
        ),
        # If you need a mix of both,
        # you should rewrite the top-down one.
        # Thus, you can do things in the order that pleases you,
        # recurse when you want, etc.,
        # and you can add flags in the dicts
        # to avoid multi-post-processings
        # on recursed dicts.
        # But clearly, in most cases,
        # the bottom-up post-processing makes more sense.
        # (Think sorting sub-results, etc.)
    }
    result.update(result_content)
    return result


def init_connections_extra_data(
    get_extra_data_template_callback,
    manager=None,
    empty_stash_stack=False,
):
    """
    You should call this function or a custom one
    before using the custom query wrapper.
    """

    connections.django_monkey_patches_dict = (
        get_extra_data_template_callback()
    )
    if empty_stash_stack or not hasattr(
        connections, "django_monkey_patches_stash_stack"
    ):
        connections.django_monkey_patches_stash_stack = (
            get_managed_list(manager)
        )

    for connection_key in connections:
        connection = connections[connection_key]
        connection.django_monkey_patches_dict = (
            get_extra_data_template_callback()
        )
        if empty_stash_stack or not hasattr(
            connection, "django_monkey_patches_stash_stack"
        ):
            connection.django_monkey_patches_stash_stack = (
                get_managed_list(manager)
            )


# pylint: disable-next=too-many-arguments
def init_connections_extra_data_v1(
    manager=None,
    empty_stash_stack=False,
    query_fields=None,
    min_seconds_threshold=0,
    max_seconds_threshold=float("inf"),
    filter_callback=None,
    insertion_callback=None,
    subsets_extra_data=None,
    allocated_subsets_extra_data=None,
    allocated_subsets_key_callback=None,
    allocated_subsets_init_callback=None,
    top_down_post_processing_callback=None,
    bottom_up_post_processing_callback=None,
):
    """
    A simple default function providing the
    get_extra_data_template_callback argument
    to init_connections_extra_data().
    """

    def lambda_get_extra_data_template_for_set_of_queries_v1():
        return get_extra_data_template_for_set_of_queries_v1(
            query_fields=query_fields,
            min_seconds_threshold=min_seconds_threshold,
            max_seconds_threshold=max_seconds_threshold,
            filter_callback=filter_callback,
            insertion_callback=insertion_callback,
            subsets_extra_data=subsets_extra_data,
            allocated_subsets_extra_data=allocated_subsets_extra_data,
            allocated_subsets_key_callback=(
                allocated_subsets_key_callback
            ),
            allocated_subsets_init_callback=(
                allocated_subsets_init_callback
            ),
            top_down_post_processing_callback=(
                top_down_post_processing_callback
            ),
            bottom_up_post_processing_callback=(
                bottom_up_post_processing_callback
            ),
            manager=manager,
        )

    init_connections_extra_data(
        lambda_get_extra_data_template_for_set_of_queries_v1,
        manager=manager,
        empty_stash_stack=empty_stash_stack,
    )


# pylint: disable-next=too-many-arguments
def insert_in_connections_extra_data_v1(
    execute,
    sql,
    params,
    many,
    context,
    call_stack,
    start_time,
    result,
    end_time,
    duration,
):
    """
    The entry-point function to insert query data in relevant dicts.
    It should be your default POST_EXECUTION_CALLBACK.
    """

    all_dicts = [
        connections.django_monkey_patches_dict,
        context["connection"].django_monkey_patches_dict,
    ]
    data = {
        "execute": execute,
        "sql": sql,
        "params": params,
        "many": many,
        "context": context,
        "call_stack": call_stack,
        "start_time": start_time,
        "result": result,
        "end_time": end_time,
        "duration": duration,
    }
    for extra_data_dict in all_dicts:
        insert_in_extra_data_dict_v1(extra_data_dict, data)
    return result


# pylint: disable-next=too-many-branches
def insert_in_extra_data_dict_v1(extra_data_dict, data):
    """
    The recursive function used to insert the data of a query
    in all relevant dicts.
    """

    # Filtering part -------------------------------------------------
    if data["duration"] is not None:
        # /!\ check list: you did activate TIME_QUERIES?
        if (
            extra_data_dict["min_seconds_threshold"]
            > data["duration"]
        ):
            return
        if (
            extra_data_dict["max_seconds_threshold"]
            < data["duration"]
        ):
            return

    filter_callback = extra_data_dict["filter_callback"]
    if filter_callback is not None and not filter_callback(
        extra_data_dict, data
    ):
        return
    # ----------------------------------------------------------------

    # Insertion part -------------------------------------------------
    extra_data_dict["query_count"] += 1
    fields = extra_data_dict["query_fields"]
    if len(fields) > 0:
        local_data = {}
        for field in fields:
            # pylint: disable-next=unidiomatic-typecheck
            if type(field) is str:
                local_data[field] = data[field]
            else:
                # It is a function.
                local_data[field.field_name] = field(
                    extra_data_dict, data
                )
        extra_data_dict["query_list"].append(local_data)
    if data["duration"] is not None:
        extra_data_dict["total_duration"] += data["duration"]
        extra_data_dict["min_duration"] = min(
            extra_data_dict["min_duration"], data["duration"]
        )
        extra_data_dict["max_duration"] = max(
            extra_data_dict["max_duration"], data["duration"]
        )
    insertion_callback = extra_data_dict["insertion_callback"]
    if insertion_callback is not None:
        # You can keep track of nesting
        # by altering data in this callback.
        insertion_callback(extra_data_dict, data)
    # ----------------------------------------------------------------

    # Recursion part -------------------------------------------------
    for some_dict in extra_data_dict["subsets_extra_data"]:
        insert_in_extra_data_dict_v1(some_dict, data)
    for (
        allocated_subset_key,
        key_generator,
    ) in extra_data_dict["allocated_subsets_key_callback"].items():
        if key_generator is None:
            continue
        sub_dict = extra_data_dict["allocated_subsets_extra_data"][
            allocated_subset_key
        ]
        if sub_dict is None:
            raise ValueError(
                "allocated_subsets error:"
                f" no sub_dict for {allocated_subset_key}"
            )
        init_callback = extra_data_dict[
            "allocated_subsets_init_callback"
        ][allocated_subset_key]
        if init_callback is None:
            raise ValueError(
                "allocated_subsets error:"
                f" no init_callback for {allocated_subset_key}"
            )
        some_key = key_generator(extra_data_dict, data)
        if sub_dict.get(some_key) is None:
            sub_dict[some_key] = init_callback(extra_data_dict, data)
        insert_in_extra_data_dict_v1(sub_dict[some_key], data)
    # ----------------------------------------------------------------


def synthetize_connections_extra_data_v1():
    """
    The entry-point function to call synthetize_extra_data_dict_v1()
    on all root extra data dicts.
    """

    extra_data_dicts = [connections.django_monkey_patches_dict]
    extra_data_dicts.extend(
        connections[connection_key].django_monkey_patches_dict
        for connection_key in connections
    )
    for extra_data_dict in extra_data_dicts:
        synthetize_extra_data_dict_v1(extra_data_dict)


def synthetize_extra_data_dict_v1(extra_data_dict):
    """
    The recursive function used to synthetize
    the results at the end.
    """

    if extra_data_dict["query_count"] > 0:
        extra_data_dict["average_duration"] = (
            extra_data_dict["total_duration"]
            / extra_data_dict["query_count"]
        )

    # Top-down
    processing_callback = extra_data_dict[
        "top_down_post_processing_callback"
    ]
    if processing_callback is not None:
        processing_callback(extra_data_dict)

    # Recursion
    for some_dict in extra_data_dict["subsets_extra_data"].values():
        synthetize_extra_data_dict_v1(some_dict)
    for sub_dict in extra_data_dict[
        "allocated_subsets_extra_data"
    ].values():
        for some_dict in sub_dict.values():
            synthetize_extra_data_dict_v1(some_dict)

    # Bottom-up
    processing_callback = extra_data_dict[
        "bottom_up_post_processing_callback"
    ]
    if processing_callback is not None:
        processing_callback(extra_data_dict)


def reorder_dict_by_total_duration_of_sub_dicts(some_dict):
    """
    A common tool to visualize the result of profiling DB queries
    is to see at the top
    similar queries where most of the time is spent.
    """

    couples_list = list(some_dict.items())
    couples_list.sort(
        key=lambda x: x[1]["total_duration"],
        reverse=True,
    )
    return dict(couples_list)


def apply_reorder_dict_by_total_duration_of_sub_dicts_to(
    main_dict,
    some_sub_dict_key,
):
    """
    This function simplifies use of the previous one.
    """

    main_dict[some_sub_dict_key] = (
        reorder_dict_by_total_duration_of_sub_dicts(
            main_dict[some_sub_dict_key]
        )
    )


def wrap_connections(
    exit_stack,
    custom_query_wrapper,
    connections_to_wrap=None,
):
    """
    Apply the context-manager connection.execute_wrapper()
    with the given custom_query_wrapper
    on an ExitStack() instance,
    for given connections or all connections if none is given.
    """

    if connections_to_wrap is None:
        connections_to_wrap = connections
    for connection_key in connections_to_wrap:
        connection = connections[connection_key]
        exit_stack.enter_context(
            connection.execute_wrapper(custom_query_wrapper)
        )


def stash_extra_data_dicts(reinit_after_stash=None):
    """
    Stash current django_monkey_patches_dicts
    in django_monkey_patches_stash_stacks.
    """

    connections.django_monkey_patches_stash_stack.append(
        connections.django_monkey_patches_dict
    )

    if reinit_after_stash is None:
        connections.django_monkey_patches_dict = None
    for connection_key in connections:
        connection = connections[connection_key]
        connection.django_monkey_patches_stash_stack.append(
            connection.django_monkey_patches_dict
        )
        if reinit_after_stash is None:
            connection.django_monkey_patches_dict = None

    if reinit_after_stash:
        reinit_after_stash()


# pylint: disable-next=too-many-arguments
def stash_extra_data_dicts_and_reinit_v1(
    manager=None,
    query_fields=None,
    min_seconds_threshold=0,
    max_seconds_threshold=float("inf"),
    filter_callback=None,
    insertion_callback=None,
    subsets_extra_data=None,
    allocated_subsets_extra_data=None,
    allocated_subsets_key_callback=None,
    allocated_subsets_init_callback=None,
    top_down_post_processing_callback=None,
    bottom_up_post_processing_callback=None,
):
    """
    Stash current django_monkey_patches_dicts
    in django_monkey_patches_stash_stacks,
    and apply a reinit using default init functions.
    """

    def lambda_init_connections_extra_data_v1():
        init_connections_extra_data_v1(
            manager=manager,
            empty_stash_stack=False,
            query_fields=query_fields,
            min_seconds_threshold=min_seconds_threshold,
            max_seconds_threshold=max_seconds_threshold,
            filter_callback=filter_callback,
            insertion_callback=insertion_callback,
            subsets_extra_data=subsets_extra_data,
            allocated_subsets_extra_data=allocated_subsets_extra_data,
            allocated_subsets_key_callback=(
                allocated_subsets_key_callback
            ),
            allocated_subsets_init_callback=(
                allocated_subsets_init_callback
            ),
            top_down_post_processing_callback=(
                top_down_post_processing_callback
            ),
            bottom_up_post_processing_callback=(
                bottom_up_post_processing_callback
            ),
        )

    stash_extra_data_dicts(
        reinit_after_stash=lambda_init_connections_extra_data_v1
    )


def pop_extra_data_dicts():
    """
    Pop last dicts in django_monkey_patches_stash_stacks
    to django_monkey_patches_dicts.
    """

    connections.django_monkey_patches_dict = (
        connections.django_monkey_patches_stash_stack.pop()
    )
    for connection_key in connections:
        connection = connections[connection_key]
        connection.django_monkey_patches_dict = (
            connection.django_monkey_patches_stash_stack.pop()
        )


# You should extract the data to logs or files,
# during execution or at the end,
# using custom insertion_callback or post_processing_callback.
# Here is an example of intermediate complexity
# using all of the above.
#
# from contextlib import ExitStack
#
# from django_monkey_patches import django__query_wrapper
# from django_monkey_patches.django__query_wrapper import (
#     apply_reorder_dict_by_total_duration_of_sub_dicts_to,
#     custom_query_wrapper_v1,
#     get_extra_data_template_for_set_of_queries_v1,
#     get_sql_signature_v1,
#     init_connections_extra_data_v1,
#     insert_in_connections_extra_data_v1,
#     synthetize_connections_extra_data_v1,
# )
#
# # /!\ Beware of double counting that may occur when you use
# # COUNT_QUERIES and insert_in_connections_extra_data_v1
# # simultaneously.
# # If you want to monitor COUNT_QUERIES all the time,
# # and activate something else more rarely,
# # you have 2 possibilities : unactivate COUNT_QUERIES when needed,
# # or activate insert_in_connections_extra_data_v1
# # with an insertion_callback that will cancel the double counting.
# # django__query_wrapper.COUNT_QUERIES = True
# django__query_wrapper.TIME_QUERIES = True
# django__query_wrapper.COMPUTE_CALL_STACK = True
# django__query_wrapper.POST_EXECUTION_CALLBACK = (
#     insert_in_connections_extra_data_v1
# )
#
#
# class CustomQueryWrapperMiddleware:
#     def __init__(self, get_response):
#         self.get_response = get_response
#
#     def __call__(self, request):
#         init_connections_extra_data_v1()
#         connection.django_monkey_patches_dict[
#             "allocated_subsets_extra_data"
#         ] = {"per_sql_signature_v1": {},}
#         connection.django_monkey_patches_dict[
#             "allocated_subsets_key_callback"
#         ] = {"per_sql_signature_v1": get_sql_signature_v1,}
#         connection.django_monkey_patches_dict[
#             "allocated_subsets_init_callback"
#         ] = {
#             "per_sql_signature_v1": (
#                 lambda x, y: (
#                     get_extra_data_template_for_set_of_queries_v1(
#                         query_fields=[
#                             "sql",
#                             "params",
#                             "many",
#                             # "call_stack",
#                             get_light_call_stack_v1,
#                             "start_time",
#                             "result",
#                             "end_time",
#                             "duration",
#                             get_full_query_v1,
#                         ],
#                     )
#                 )
#             ),
#         }
#
#         connection.django_monkey_patches_dict[
#             "bottom_up_post_processing_callback"
#         ] = (
#             lambda x: (
#               apply_reorder_dict_by_total_duration_of_sub_dicts_to(
#                 x["allocated_subsets_extra_data"],
#                 "per_sql_signature_v1",
#               )
#             )
#         )
#         with ExitStack() as exit_stack:
#             wrap_connections(exit_stack, custom_query_wrapper_v1)
#             response = self.get_response(request)
#             synthetize_connections_extra_data_v1()
#             # dump to file or log:
#             # connections.django_monkey_patches_dict
#             # for connection_key in connections:
#             #     some_connection = connections[connection_key]
#             #     # dump to file or log:
#             #     # some_connection.django_monkey_patches_dict
#             return response



"""
This file is part of django-monkey-patches library.

django-monkey-patches is free software:
you can redistribute it and/or modify it under the terms
of the GNU Lesser General Public License
as published by the Free Software Foundation,
either version 3 of the License,
or (at your option) any later version.

django-monkey-patches is distributed in the hope
that it will be useful,
but WITHOUT ANY WARRANTY;
without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU Lesser General Public License for more details.

You should have received a copy of
the GNU Lesser General Public License
along with django-monkey-patches.
If not, see <http://www.gnu.org/licenses/>.

©Copyright 2023-2024 Laurent Lyaudet
----------------------------------------------------------------------
Django Rest Framework is considered "feature-complete".
But if you had the chance to work with an in-house framework
like I did in PHP for 8,5 years,
you know that there is no such thing as
a "feature-complete" framework :(.
The goal of this patch is to have a sensible place
to add unconventional prefetches.
You may apply the patch,
or use patched_to_representation_v1
in a class inheriting from ListSerializer.
"""

# pylint: disable-next=import-error
from django.db import models

# pylint: disable-next=import-error
from rest_framework.serializers import ListSerializer

original_to_representation = ListSerializer.to_representation


def patched_to_representation_v1(self, data):
    """
    List of object instances -> List of dicts of primitive datatypes.
    """
    # Dealing with nested relationships, data can be a Manager,
    # so, first get a queryset from the Manager if needed
    iterable = (
        data.all()
        if isinstance(data, models.manager.BaseManager)
        else data
    )

    if self.context.get("list_before_representation_callback"):
        self.context.get("list_before_representation_callback")(
            iterable
        )
    elif self.child.context.get(
        "list_before_representation_callback"
    ):
        self.child.context.get("list_before_representation_callback")(
            iterable
        )
    elif hasattr(self, "list_before_representation_callback"):
        self.list_before_representation_callback(iterable)
    elif hasattr(self.child, "list_before_representation_callback"):
        self.child.list_before_representation_callback(iterable)

    return [self.child.to_representation(item) for item in iterable]


def apply_patched_to_representation_v1():
    """
    Apply the patch to enable custom computation before the end of
    the list serializer to_representation().
    """
    ListSerializer.to_representation = patched_to_representation_v1
    return original_to_representation
